/*
    tcl-fuse - Tcl FUSE Interface
    A Tcl interface to the Linux kernel's FUSE subsystem.

    Copyright (C) 2008  Alexandros Stergiakis

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 * proxy.c
 *
 */

// TODO: Verify that the number type conversions in this file cannot overflow.
//       Search for "Int" & "Long" strings to find all the converstions.

#include <string.h> // memset, strcpy, strlen, memcpy
#include <sys/types.h> // struct stat
#include <sys/stat.h> // struct stat
#include <unistd.h> // struct stat
#include <errno.h> // error codes

#include <tcl.h>
#include <fuse.h>

#include "tclfuse.h"
#include "convert.h"

int _getattr(const char *path, struct stat *st) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    memset(st, 0, sizeof(struct stat));

    Tcl_Obj *newObjv[4];
    newObjv[argc++] = data->getattr;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);
    Tcl_Obj *result = Tcl_GetObjResult(interp);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    } else {
        if (! rconvert__stat(interp, st, result)) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
        // else: success
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

// TODO: Support the incremental form of the FUSE operation 
//       Read about it in "fuse.h".
int _readdir(const char *path, void *buf, fuse_fill_dir_t filler,
                    off_t offset, struct fuse_file_info *fi) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[5];
    newObjv[argc++] = data->readdir;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = convert__fuse_file_info(fi);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);
    Tcl_Obj *result = Tcl_GetObjResult(interp);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    } else {
        Tcl_Obj **objv;
        int objc;
        if (Tcl_ListObjGetElements(interp, result, &objc, &objv) == TCL_ERROR) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        } else {
            int a;
            for (a = 0; a < objc; a++) {
                if (filler(buf, Tcl_GetString(objv[a]), NULL, 0) == 1) {
                    res = -ENOENT;
                    Tcl_SetResult(interp, "Buffer is full", TCL_STATIC);
                    Tcl_BackgroundError(interp);
                    break;
                }
            }
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

/* Taken from TclX */
int lempty(Tcl_Obj *obj) {
    if ((obj->typePtr == NULL) && (obj->bytes == NULL)) {
        return 1;
    }

    int length = 1;
    Tcl_ListObjLength(NULL, obj, &length);

    return (0 == length);
}

int _open(const char *path, struct fuse_file_info *fi) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[5];
    newObjv[argc++] = data->open;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = convert__fuse_file_info(fi);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);
    Tcl_Obj *result = Tcl_GetObjResult(interp);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    } else {
        int fd;
        if (! lempty(result)) {
            if (Tcl_GetIntFromObj(interp, result, &fd) == TCL_ERROR) {
                res = -ENOENT;
                Tcl_BackgroundError(interp); // error message is set.
            } else {
                fi->fh = fd;
            }
        }
    }

    for(i = 0; i < argc; i++)
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _read(const char *path, char *buf, size_t size, off_t offset, struct fuse_file_info *fi) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[7];
    newObjv[argc++] = data->read;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = convert__fuse_file_info(fi);
    newObjv[argc++] = Tcl_NewLongObj(size);
    newObjv[argc++] = Tcl_NewLongObj(offset);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);
    Tcl_Obj *result = Tcl_GetObjResult(interp);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    } else {
        unsigned char *data;
        data = Tcl_GetByteArrayFromObj(result, &res);
        memcpy(buf, data, res);
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _readlink(const char *path, char *buf, size_t size) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[4];
    newObjv[argc++] = data->readlink;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);
    Tcl_Obj *result = Tcl_GetObjResult(interp);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    } else {
        // The following works also for an empty list as result.
        char *str = Tcl_GetString(result);
        int len = strlen(str) + 1;
        if (size < len) {
            strncpy(buf, str, size - 1);
            buf[size - 1] = '\0';
        } else {
            strcpy(buf, str);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _rename(const char *from, const char *to) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[5];
    newObjv[argc++] = data->rename;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(from, -1);
    newObjv[argc++] = Tcl_NewStringObj(to, -1);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _mknod(const char *path, mode_t mode, dev_t rdev) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[7];
    newObjv[argc++] = data->mknod;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);

    /* no need to check for NULL (type not found), as 
     * there should always be a match (guaranteed
     * by FUSE)
     */
    newObjv[argc++] = convert__type(mode); // type

    /* retain only the permission bits */
    newObjv[argc++] = Tcl_NewIntObj((int) mode & 07777); // perm

    /* convert dev to major/minor pair.
     * ref: http://www.ussg.iu.edu/hypermail/linux/kernel/0103.3/0038.html
     */
    int major, minor;
    if (mode & (S_IFCHR | S_IFBLK)) { // neither of the flags is zero.
        if ((major = (rdev >> 32)) != 0) 
            minor = (rdev & 0xffffffff); 
        else if ((major = (rdev >> 16)) != 0) 
            minor = (rdev & 0xffff); 
        else { 
            major = (rdev >> 8); 
            minor = (rdev & 0xff); 
        } 
        Tcl_Obj *objv[2];
        objv[0] = Tcl_NewIntObj((int) major);
        objv[1] = Tcl_NewIntObj((int) minor);
        newObjv[argc++] = Tcl_NewListObj(2, objv); // dev
    } else {
        newObjv[argc++] = Tcl_NewListObj(0, NULL); // dev
    }

    for(i = 0; i < argc; i++)
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _mkdir(const char *path, mode_t mode) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[5];
    newObjv[argc++] = data->mkdir;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = Tcl_NewIntObj((int) mode);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _unlink(const char *path) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[4];
    newObjv[argc++] = data->unlink;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _rmdir(const char *path) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[4];
    newObjv[argc++] = data->rmdir;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _symlink(const char *from, const char *to) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[5];
    newObjv[argc++] = data->symlink;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(from, -1);
    newObjv[argc++] = Tcl_NewStringObj(to, -1);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _link(const char *from, const char *to) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[5];
    newObjv[argc++] = data->link;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(from, -1);
    newObjv[argc++] = Tcl_NewStringObj(to, -1);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _chmod(const char *path, mode_t mode) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[5];
    newObjv[argc++] = data->chmod;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = Tcl_NewIntObj((int) mode);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _chown(const char *path, uid_t uid, gid_t gid) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[6];
    newObjv[argc++] = data->chown;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = Tcl_NewIntObj((int) uid);
    newObjv[argc++] = Tcl_NewIntObj((int) gid);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _truncate(const char *path, off_t size) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[5];
    newObjv[argc++] = data->truncate;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = Tcl_NewIntObj((int) size);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _access(const char *path, int mask) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[5];
    newObjv[argc++] = data->access;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = convert__access_mask(mask);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _release(const char *path, struct fuse_file_info *fi) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[5];
    newObjv[argc++] = data->release;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = convert__fuse_file_info(fi);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    } else {
        fi->fh = 0;
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _flush(const char *path, struct fuse_file_info *fi) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[5];
    newObjv[argc++] = data->flush;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = convert__fuse_file_info(fi);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _fsync(const char *path, int datasync, struct fuse_file_info *fi) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[6];
    newObjv[argc++] = data->fsync;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = convert__fuse_file_info(fi);
    newObjv[argc++] = Tcl_NewIntObj(datasync);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _utimens(const char *path, const struct timespec tv[2]) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[6];
    newObjv[argc++] = data->utimens;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = Tcl_NewIntObj(tv[0].tv_sec); // atime
    newObjv[argc++] = Tcl_NewIntObj(tv[1].tv_sec); // mtime

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _write(const char *path, const char *buf, size_t size, off_t offset, struct fuse_file_info *fi) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[7];
    newObjv[argc++] = data->write;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = convert__fuse_file_info(fi);
    newObjv[argc++] = Tcl_NewByteArrayObj((unsigned char *) buf, (int) size);
    newObjv[argc++] = Tcl_NewIntObj(offset);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);
    Tcl_Obj *result = Tcl_GetObjResult(interp);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    } else {
        if (Tcl_GetIntFromObj(interp, result, &res) == TCL_ERROR) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _statfs(const char *path, struct statvfs *stbuf) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[4];
    newObjv[argc++] = data->statfs;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);
    Tcl_Obj *result = Tcl_GetObjResult(interp);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    } else {
        if (rconvert__statvfs(interp, stbuf, result) == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

// TODO: Make use of "flags".
int _setxattr(const char *path, const char *name, const char *value, size_t size, int flags) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[6];
    newObjv[argc++] = data->setxattr;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = Tcl_NewStringObj(name, -1);
    newObjv[argc++] = Tcl_NewByteArrayObj((unsigned char *) value, (int) size);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _getxattr(const char *path, const char *name, char *value, size_t size) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[6];
    newObjv[argc++] = data->getxattr;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = Tcl_NewStringObj(name, -1);
    newObjv[argc++] = Tcl_NewIntObj(size);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);
    Tcl_Obj *result = Tcl_GetObjResult(interp);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    } else {
        unsigned char *data;
        data = Tcl_GetByteArrayFromObj(result, &res);
        memcpy(value, data, res);
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _listxattr(const char *path, char *list, size_t size) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[5];
    newObjv[argc++] = data->listxattr;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = Tcl_NewIntObj(size);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);
    Tcl_Obj *result = Tcl_GetObjResult(interp);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    } else {
        Tcl_Obj **objv;
        int objc;
        if (Tcl_ListObjGetElements(interp, result, &objc, &objv) == TCL_ERROR) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        } else {
            char *str;
            int str_len;
            for(i = 0; i < objc; i++) {
                str = Tcl_GetStringFromObj(objv[i], &str_len);
                memcpy(list, str, str_len + 1); // XXX is \0 char counted in str_len above? We assume that it doesn't.
                list += str_len + 1; // XXX Same as above
            }
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _removexattr(const char *path, const char *name) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[5];
    newObjv[argc++] = data->removexattr;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = Tcl_NewStringObj(name, -1);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _opendir(const char *path, struct fuse_file_info *fi) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[5];
    newObjv[argc++] = data->opendir;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = convert__fuse_file_info(fi);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);
    Tcl_Obj *result = Tcl_GetObjResult(interp);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    } else {
        int fd;
        if (! lempty(result)) {
            if (Tcl_GetIntFromObj(interp, result, &fd) == TCL_ERROR) {
                res = -ENOENT;
                Tcl_BackgroundError(interp); // error message is set.
            } else {
                fi->fh = fd;
            }
        }
    }

    return res;
}

int _releasedir(const char *path, struct fuse_file_info *fi) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[5];
    newObjv[argc++] = data->releasedir;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = convert__fuse_file_info(fi);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    } else {
        fi->fh = 0;
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _fsyncdir(const char *path, int datasync, struct fuse_file_info *fi) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[6];
    newObjv[argc++] = data->fsyncdir;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = convert__fuse_file_info(fi);
    newObjv[argc++] = Tcl_NewIntObj(datasync);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _create(const char *path, mode_t mode, struct fuse_file_info *fi) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[6];
    newObjv[argc++] = data->create;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = convert__fuse_file_info(fi);
    newObjv[argc++] = convert__type(mode);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);
    Tcl_Obj *result = Tcl_GetObjResult(interp);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    } else {
        int fd;
        if (! lempty(result)) {
            if (Tcl_GetIntFromObj(interp, result, &fd) == TCL_ERROR) {
                res = -ENOENT;
                Tcl_BackgroundError(interp); // error message is set.
            } else {
                fi->fh = fd;
            }
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _ftruncate(const char *path, off_t size, struct fuse_file_info *fi) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[6];
    newObjv[argc++] = data->ftruncate;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = convert__fuse_file_info(fi);
    newObjv[argc++] = Tcl_NewIntObj((int) size);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _fgetattr(const char *path, struct stat *st, struct fuse_file_info *fi) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    memset(st, 0, sizeof(struct stat));

    Tcl_Obj *newObjv[5];
    newObjv[argc++] = data->fgetattr;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = convert__fuse_file_info(fi);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);
    Tcl_Obj *result = Tcl_GetObjResult(interp);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    } else {
        if (! rconvert__stat(interp, st, result)) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
        // else: success
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

void _destroy(void *user_data) {
    int i, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[3];
    newObjv[argc++] = data->destroy;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        if (convert__errno(interp) == 0) {
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);
}

int _bmap(const char *path, size_t blocksize, uint64_t *idx) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[6];
    newObjv[argc++] = data->bmap;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = Tcl_NewIntObj((int) blocksize);
    newObjv[argc++] = Tcl_NewLongObj((long) idx);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}

int _lock(const char *path, struct fuse_file_info *fi, int cmd, struct flock *lock) {
    int i, res = 0, argc = 0;

    struct tclfuse_fsdata *data = (struct tclfuse_fsdata *) fuse_get_context()->private_data;
    Tcl_Interp *interp = data->interp;

    Tcl_Obj *newObjv[7];
    newObjv[argc++] = data->lock;
    if (data->userdata != NULL) {
        newObjv[argc++] = data->userdata;
    }
    newObjv[argc++] = convert__fuse_context(fuse_get_context());
    newObjv[argc++] = Tcl_NewStringObj(path, -1);
    newObjv[argc++] = convert__fuse_file_info(fi);
    newObjv[argc++] = convert__lockcmd(cmd);
    newObjv[argc++] = convert__flock(lock);

    for(i = 0; i < argc; i++) 
        Tcl_IncrRefCount(newObjv[i]);

    int code = Tcl_EvalObjv(interp, argc, newObjv, TCL_EVAL_GLOBAL);
    Tcl_Obj *result = Tcl_GetObjResult(interp);

    if (code == TCL_ERROR) {
        res = convert__errno(interp);
        if (res == 0) {
            res = -ENOENT;
            Tcl_BackgroundError(interp);
        }
    } else {
        if (! lempty(result)) {
            int pid;
            if (Tcl_GetIntFromObj(interp, result, &pid) == TCL_ERROR) {
                res = -ENOENT;
                Tcl_BackgroundError(interp); // error message is set.
            } else {
                lock->l_pid = pid;
            }
        }
    }

    for(i = 0; i < argc; i++) 
        Tcl_DecrRefCount(newObjv[i]);

    return res;
}
