#!/usr/bin/env tclsh

package provide tktopdf 0.5

package require Tk
package require Tcl 8.5
package require tktocanvas

#source tktocanvas.tcl
#source base64.tcl
#source transformations.tcl

namespace eval tktopdf {

# variables used when GetWindowConfig is called. they values are
# coordinates of parent window.

set parentx 0
set parenty 0

proc GetWindowConfig {win} {

    set winChildren [winfo children $win]

    if {$win != "."} {
        set class [winfo class $win]
    } else {
        set class "Toplevel"
    }

    set config [eval "Get${class}Config $win"]
    set childrenList {}
    
    set tktopdf::parentx [dict get $config -x]
    set tktopdf::parenty [dict get $config -y]
    
    if {($class == "Toplevel") && ([$win cget -menu] != "")} {
        # Get menu clone's name.
        
        set menuName [string map {. #} [$win cget -menu]]
        set menuName [lsearch -regexp $winChildren $menuName]
        set menuName [lindex [winfo children $win] $menuName]
       
        lappend childrenList [GetMenuConfig $menuName]
      
        set winChildren [lreplace $winChildren \
                [lsearch $winChildren $menuName] \
                [lsearch $winChildren $menuName]]
        
        incr tktopdf::parenty [winfo height $menuName]
    }

    foreach childWindow $winChildren {

        if {![winfo ismapped $childWindow]} {
            continue
        }
        
        set class [winfo class $childWindow]
        
        if {$class == "Toplevel"} {
            continue
        }
        
        if {([llength [winfo children $childWindow]] != 0) \
                || (($class == "Frame") || ($class == "Labelframe") \
                && [$childWindow cget -container])} {
            
            if {(($class == "Labelframe") || ($class == "Frame")) \
                    && [$childWindow cget -container]} {
                set childWindow [GetEmbeddedToplevel . [winfo id $childWindow]]
                #puts $childWindow
            }

            # save global values before recursion.
            
            set backupx $tktopdf::parentx
            set backupy $tktopdf::parenty
           
            set childrenConfig [GetWindowConfig $childWindow] 
            if {$childrenConfig != 0} { 
                lappend childrenList $childrenConfig
            }

            set tktopdf::parentx $backupx
            set tktopdf::parenty $backupy
            
            continue 
        }
        
        set cmd "Get${class}Config $childWindow"

        set childConfig [eval $cmd]
        if {$childConfig != 0} { 
            lappend childrenList $childConfig
        }
    }

    if {$childrenList != {}} {
        dict set config -child $childrenList
    }

    return $config
}

proc GetEmbeddedToplevel {parent containerId} {

    set result ""

    foreach win [winfo children $parent] {
        if {([winfo class $win] == "Toplevel") \
                && ([$win cget -use] == $containerId)} {
            return $win
        }

        if {[llength [winfo children $win]] > 0} {
            set tmpResult [GetEmbeddedToplevel $win $containerId]
            if {$tmpResult != ""} {
                return $tmpResult
            }
        }
    }

    return $result
}

#############################################################################    
# proc GetWidgetStdConfig -- 
#
#       get basic configuration values from widget, and do some computations.
#
# arguments:
#       window - a window where the widget is.
#       config - a dictionary where gained information would be stored.
#
#       this procedure also using global in tktopdf package variables
#       parentx and parenty which are coordinates of parent window 
#       (do not miss with coordinates in parent).
#
# results:
#       a config dictionary will be fill with basic configuration values,
#       which are common for every widget. there is also do some computations
#       like computate 3d shadow colors of widget's border

proc GetWidgetStdConfig {window configName} {

    upvar $configName config
    
    foreach val [$window configure] {
        if {[llength $val] != 2} {
            dict set config [lindex $val 0] [lindex $val end]
        }
    }

    dict set config -name $window
    
    # if there is highlightthickness option the button will be small
    # hFrame pixels
    
    if {[dict exists $config -highlightthickness] && \
            [dict get $config -highlightthickness]} {
        set hFrame [dict get $config -highlightthickness]
    } else {
        set hFrame 0
    }

    dict set config -x [expr {[winfo x $window] + $hFrame + $tktopdf::parentx}]
    dict set config -y [expr {[winfo y $window] + $hFrame + $tktopdf::parenty}]
    
    dict set config -class [winfo class $window]

    if {$window == "."} {
        dict set config -class "Toplevel"
    }

	dict set config -height [expr {[winfo height $window] - 2 * $hFrame}]
    dict set config -width [expr {[winfo width $window] - 2 * $hFrame}]

    # If toplevel has -use option set, use it's container window coordinates.

    if {[dict get $config -class] == "Toplevel" \
            && [dict get $config -use] != ""} {
        dict set config -height [winfo height \
                [winfo pathname [dict get $config -use]]]
        dict set config -width [winfo width \
                [winfo pathname [dict get $config -use]]]
        dict set config -x [expr {\
                [winfo x [winfo pathname [dict get $config -use]]] \
                + $tktopdf::parentx}]
        dict set config -y [expr { \
                [winfo y [winfo pathname [dict get $config -use]]] \
                + $tktopdf::parenty}]
    }
    
    # compute (x1, y1) coordinates from length and width
    
    dict set config -x1 [expr {[dict get $config -x] \
            + [dict get $config -width]-1}]
    dict set config -y1 [expr {[dict get $config -y] \
            + [dict get $config -height]-1}]

    # compute colors of borders

    set shadows [ComputeShadowColors $window [dict get $config -relief] \
            [dict get $config -background]]
    dict set config -darkshadow [lindex $shadows 0]
    dict set config -lightshadow [lindex $shadows 1]

    # there is nothing more to do if the widget's class is menu

    if {[dict get $config -class] == "Menu"} {
        return
    }

    # check if widget currently is focused
    
    dict set config -hasfocus [HasFocus $window]

    # compute colors to draw 
   
    set colors [SetColorsToDraw $config]
    dict set config -foreground [lindex $colors 0]
    dict set config -background [lindex $colors 1]
    dict set config -highlightbackground [lindex $colors 2]
}

proc GetToplevelConfig {tl} {
    
    set config [dict create]
    GetWidgetStdConfig $tl config

    return $config
}

proc GetPanedwindowConfig {tl} {
    
    set config [dict create]
    GetWidgetStdConfig $tl config

    return $config
}

#############################################################################
# proc GetButtonConfig -- 
# 
#       get and compute configuration values specific to buttons. also label 
#       widget uses this procedure.
#
# arguments:
#       btn - a widget's path.
# results:
#       after this procedure is called, config dictionary will have all
#       information which is needed to export widget's representation to pdf.

proc GetButtonConfig {btn} {

    set config [dict create]

    # get configuration values
    
    GetWidgetStdConfig $btn config

    # change padx/pady dimension to pixel dimension
    
    dict set config -padx [winfo pixels $btn [dict get $config -padx]]
    dict set config -pady [winfo pixels $btn [dict get $config -pady]]

    dict set config -havetext 0
    dict set config -haveimage 0

    # check if there are image/text
    
    if {[dict get $config -image] != ""} {
        dict set config -haveimage 1
        dict set config -imageheight [image height [dict get $config -image]]
        dict set config -imagewidth [image width [dict get $config -image]]
    }

    if {([dict get $config -text] != "") \
            && (!(([dict get $config -image] != "") \
            && ([dict get $config -compound] == "none")))} {
        
        dict set config -havetext 1
        
        set textLayout [ComputeTextLayout [dict get $config -text] \
                [dict get $config -font] [dict get $config -wraplength]]
        dict set config -textwidth [lindex $textLayout 0]
        dict set config -textheight [lindex $textLayout 1]

        if {[dict get $config -wraplength]} {
            dict set config -text [lindex $textLayout 2]
        }
    }

    # if this is button iSpace and iDiam are set to 0'es
    
    set indicator [ComputeIndicatorSpace $config]
    # set indicator space
    
    dict set config -ispace [lindex $indicator 0]
    # set indicator diameter
    
    dict set config -idiam [lindex $indicator 1]

    #
    # set the font and image positions in widget
    #
   
    if {[dict get $config -haveimage] && [dict get $config -havetext] \
            && [dict get $config -compound] != ""} {

        set coords [BtnGetTextAndImageCoords $config]
        dict set config -textcoords [lindex $coords 0]
        dict set config -imagecoords [lindex $coords 1]

        foreach {coords width height name data ddd} [GetImageData \
                [dict get $config -imagecoords] [dict get $config -image] \
                "nw"] {
            dict set config -imagecoords $coords
            dict set config -imagename $name
            dict set config -data $data
        }
    } elseif {[dict get $config -haveimage] != 0} {
        # if button has an image
        
        dict set config -imagecoords [BtnGetImageCoords $config \
            [expr {[dict get $config -ispace]+[dict get $config -imagewidth]}]\
            [dict get $config -imageheight] 0 0 0 0]
        
        foreach {coords width height name data ddd} [GetImageData \
                [dict get $config -imagecoords] [dict get $config -image] \
                "nw"] {
            dict set config -imagecoords $coords
            dict set config -imagename $name
            dict set config -data $data
        }
    } elseif {[dict get $config -havetext] != 0} {
        # if button has a text
        
        dict set config -textcoords [BtnGetTextCoords $config \
            [expr {[dict get $config -ispace]+[dict get $config -textwidth]}]\
            [dict get $config -textheight] 0 0]
    }

    # do radiobutton/checkbutton specific stuff
    
    if {[dict get $config -class] == "Radiobutton" ||
        [dict get $config -class] == "Checkbutton"} {
        # check if button is checked
        
        dict set config -selected [ButtonIsSelected $btn]

        if {[dict get $config -selected] \
                && [dict get $config -selectimage] != ""} {

            dict set config -image [dict get $config -selectimage]
        }

        # compute indicator's coords, must be call after computing 
        # text/image coords
       
        set indicatorCoords [ComputeIndicatorCoords $config]
        dict set config -ix [lindex $indicatorCoords 0]
        dict set config -iy [lindex $indicatorCoords 1]
        dict set config -ix1 [lindex $indicatorCoords 2]
        dict set config -iy1 [lindex $indicatorCoords 3]

        # override the relief specified for the button if this is 
        # a checkbutton or radiobutton and there's no indicator.
        
        if {([dict get $config -class] == "Radiobutton" \
                || [dict get $config -class] == "Checkbutton") \
                && (![dict get $config -indicatoron]) \
                && [ButtonIsSelected $btn]} {
            
            if {[dict get $config -selected]} {
                
                dict set config -relief "sunken"
                dict set config -background white 
                #[dict get $config -highlightbackground]
               
                # recompute shadows color

                set shadows [ComputeShadowColors $btn \
                        [dict get $config -relief] \
                        [dict get $config -background]]
                dict set config -darkshadow [lindex $shadows 0]
                dict set config -lightshadow [lindex $shadows 1]

            } elseif {[dict get $config -overrelief] \
                    !=[dict get $config -relief]} {
            
                dict set config -relief offrelief
            }
        }
    }

    if {[dict exists $config -ix]} {
        dict set config -markcoords [ComputeMarkCoords [dict get $config -ix] \
                [dict get $config -iy] [dict get $config -ix1] \
                [dict get $config -iy1]]
    }

    return $config
}

#############################################################################
# proc GetCheckbuttonConfig --
#
#       get and compute configuration values specific to checkbutton widget.
#       this procedure is only wrapper procedure for getButtonConfig.
#
# arguments:
#       chbtn - a widget's path.
# results:
#       after this procedure is called, config dictionary will have all
#       information which is needed to export widget's representation to pdf.

proc GetCheckbuttonConfig {chbtn} {

    return [GetButtonConfig $chbtn]
}

#############################################################################
# proc GetRadiobuttonConfig --
#
#       get and compute configuration values specific to radiobutton widget.
#       this procedure is only wrapper procedure for getButtonConfig.
#
# arguments:
#       radiobtn - a widget's path.
# results:
#       after this procedure is called, config dictionary will have all
#       information which is needed to export widget's representation to pdf.

proc GetRadiobuttonConfig {radiobtn} {

    return [GetButtonConfig $radiobtn]
}

#############################################################################
# proc GetLabelConfig --
#
#       get and compute configuration values specific to label widget.
#       this procedure is only wrapper procedure for getButtonConfig.
#
# arguments:
#       lab - a widget's path.
# results:
#       after this procedure is called, config dictionary will have all
#       information which is needed to export widget's representation to pdf.

proc GetLabelConfig {lab} {

    return [GetButtonConfig $lab]
}

#############################################################################
# proc GetEntryConfig --
#
#       get and compute configuration values specific to entry widget.
#
# arguments:
#       en - a widget's path.
# results:
#       after this procedure is called, config dictionary will have all
#       information which is needed to export widget's representation to pdf.

proc GetEntryConfig {en} {

    set config [dict create]

    # get configuration values.

    GetWidgetStdConfig $en config
  
    dict set config -text [$en get]

    if {[dict get $config -class] == "Entry" \
            && [dict get $config -show] != ""} {
   
        dict set config -text [string repeat [dict get $config -show] \
                [string length [dict get $config -text]]]
    }

    set textLayout [ComputeTextLayout [dict get $config -text] \
            [dict get $config -font] 0]
    dict set config -textwidth [lindex $textLayout 0]
    dict set config -textheight [lindex $textLayout 1]

    set textCoords [$en bbox 0]
    
    dict set config -textcoords [list [expr {[dict get $config -x] \
            + [lindex $textCoords 0] \
            + ([dict get $config -textwidth] / 2) \
            - [dict get $config -highlightthickness]}] \
            [expr {[dict get $config -y] \
            + [lindex $textCoords 1] + ([dict get $config -textheight]/2) \
            - [dict get $config -highlightthickness]}]]

    if {[dict get $config -class] == "Spinbox"} {
        set halfHeight [expr {([dict get $config -y1]-[dict get $config -y] - 1)/2}]
        
        set widgetWidth [expr {[dict get $config -width] \
                + 2 * [dict get $config -highlightthickness]}]
        set i [dict get $config -borderwidth]
        for {set place [$en identify $i $halfHeight]} \
                {$place != "buttonup" && $place != "buttondown" \
                && ($i < $widgetWidth)} {incr i} {
            
            set place [$en identify $i $halfHeight]
        }
        incr i -1

        dict set config -bupx [expr {[dict get $config -x] + $i \
                - [dict get $config -highlightthickness]}]
        dict set config -bupy [expr {[dict get $config -y] \
                + [dict get $config -borderwidth]}]
        dict set config -bupx1 [expr {[dict get $config -x1] \
                - [dict get $config -borderwidth]}]
        dict set config -bupy1 [expr {[dict get $config -y] \
                + $halfHeight}]
        
        dict set config -bdownx [expr {[dict get $config -x] + $i \
                - [dict get $config -highlightthickness]}]
        dict set config -bdowny [expr {[dict get $config -y] \
                + $halfHeight + 1}]
        dict set config -bdownx1 [expr {[dict get $config -x1] \
                - [dict get $config -borderwidth]}]
        dict set config -bdowny1 [expr {[dict get $config -y1] \
                - [dict get $config -borderwidth] }]

        # Button arrows.
        
        set shadows [ComputeShadowColors $en \
                [dict get $config -buttonuprelief] \
                [dict get $config -background]]

        dict set config -upbuttonshadows [list [lindex $shadows 1] \
                [lindex $shadows 0]]
        
        set shadows [ComputeShadowColors $en \
                [dict get $config -buttondownrelief] \
                [dict get $config -background]]

        dict set config -downbuttonshadows [list [lindex $shadows 1] \
                [lindex $shadows 0]]
        
        set pad 2
        set inset [expr {[dict get $config -borderwidth] \
                + [dict get $config -highlightthickness]}]

        set xWidth [font measure [dict get $config -font] "0"]
        
        if {$xWidth < 7} {
            set xWidth 7
        }
       
        if {($xWidth % 2) == 0} {
            incr xWidth
        }

        set bIsPressed 0

        set space [expr {[dict get $config -bupy1] \
                - [dict get $config -bupy] - (2 * $pad)}]
        
        # tHeight = halfHeight 
        set tHeight [expr {($xWidth + 1) / 2}]
        if {$tHeight > $space} {
            set tHeight $space
        }
        set space [expr {($space - $tHeight) / 2}]
        
        set starty [expr {[dict get $config -bupy1] - $pad - $space}]

        set upButtonCoords [expr {[dict get $config -bupx] + $pad}]
        lappend upButtonCoords [expr {$starty - $bIsPressed}]
        
        lappend upButtonCoords [expr {[dict get $config -bupx] \
                + (($xWidth + 2)  / 2) + 1}]
        lappend upButtonCoords [expr {$starty - $tHeight - $bIsPressed}]
        
        lappend upButtonCoords [expr {[dict get $config -bupx] + $xWidth}]
        lappend upButtonCoords [lindex $upButtonCoords 1]

        dict set config -upButtonCoords $upButtonCoords
       
        set starty [expr {[dict get $config -bdowny] + $pad + $space}]
        
        set downButtonCoords [expr {[dict get $config -bupx] + $pad}]
        lappend downButtonCoords [expr {$starty - $bIsPressed}]
        
        lappend downButtonCoords [expr {[dict get $config -bupx] \
                + (($xWidth + 2)  / 2) + 1}]
        lappend downButtonCoords [expr {$starty + $tHeight + $bIsPressed}]
        
        lappend downButtonCoords [expr {[dict get $config -bupx] + $xWidth}]
        lappend downButtonCoords [lindex $downButtonCoords 1]
        
        dict set config -downButtonCoords $downButtonCoords
    }

    return $config
}

proc GetSpinboxConfig {spin} {

     return [GetEntryConfig $spin]
}

#############################################################################
# proc GetFrameConfig --
#
#       get and compute configuration values specific to frame widget.
#
# arguments:
#       frm - a widget's path.
# results:
#       after this procedure is called, config dictionary will have all
#       information which is needed to export widget's representation to pdf.

proc GetFrameConfig {frm} {

    # if this frame is container we do not draw it.

    if {[$frm cget -container]} {
        return 0
    }

    set config [dict create]

    # get configuration values.

    GetWidgetStdConfig $frm config
    
    if {[dict get $config -class] == "Labelframe"} {
        set labelConf [LabelframeGetLabelInfo $config]
        
        dict set config -shiftx [lindex $labelConf 0]
	    dict set config -shifty [lindex $labelConf 1]
        dict set config -shiftx1 [lindex $labelConf 2]
        dict set config -shifty1 [lindex $labelConf 3]
        dict set config -textcoords [lindex $labelConf 4]
        dict set config -labelboxcoords [lindex $labelConf 5]

        set textLayout [ComputeTextLayout [dict get $config -text] \
                [dict get $config -font] 0]
        dict set config -textwidth [lindex $textLayout 0]
        dict set config -textheight [lindex $textLayout 1]
    }

    return $config
}

#############################################################################
# proc GetLabelframeConfig --
#
#       get and compute configuration values specific to labelframe widget.
#       this procedure is only wrapper procedure for getFrameConfig.
#
# arguments:
#       frm - a widget's path.
# results:
#       after this procedure is called, config dictionary will have all
#       information which is needed to export widget's representation to pdf.

proc GetLabelframeConfig {labfrm} {
    
    return [GetFrameConfig $labfrm]
}

#############################################################################
# proc GetListboxConfig --
#
#       get and compute configuration values specific to listbox.
#
# arguments:
#       lbox - a widget's path.
# results:
#       after this procedure is called, config dictionary will have all
#       information which is needed to export widget's representation to pdf.

proc GetListboxConfig {lbox} {

    set config [dict create]

    GetWidgetStdConfig $lbox config

    # the index of element located at the bottom of widget.
    
    set bottomElementInd [$lbox nearest [dict get $config -y1]]
    
    set elementsInfo {}
    set elementInd [expr {$bottomElementInd - ([$lbox cget -height] - 1)}]
    set xView [$lbox xview]
    
    # get the longest element.
    set maxLen 0
    foreach el [$lbox get 0 end] {
        set len [string length $el]
        if {$len > $maxLen} {
            set maxLen $len
        }
    }

    # a list of selected elements.

    set selElements [$lbox curselection] 

    # compute a height of element in listbox, taking into account scaling.

    set elementHeight [expr {([dict get $config -height] \
            - 2 * [dict get $config -borderwidth]) / [$lbox cget -height]}]

    set tx [expr {[dict get $config -x] + [dict get $config -borderwidth]}]
    set tx1 [expr {[dict get $config -x1] - [dict get $config -borderwidth]}]
    set ty1 [expr {[dict get $config -y1] + [dict get $config -borderwidth]}]
    set ty [expr {[dict get $config -y] + [dict get $config -borderwidth] \
            - $elementHeight}]

    set ind 1
    # get info about particular elements of listbox.

    for {} {$elementInd <= $bottomElementInd} {incr elementInd} {
        
        set borders [lindex [$lbox bbox $elementInd] 0]
        
        # TODO: it needs a comment 
        if {$borders == ""} {
            continue
        } 

        set currElInfo {}
        set isSelected [lsearch $selElements $elementInd]
       
        lappend currElInfo $elementInd
        
        set text [$lbox get $elementInd]
        
        lappend currElInfo $text
        set textLayout [ComputeTextLayout $text [dict get $config -font] 0]

        set textCoords [expr {[dict get $config -x] + $borders \
                + ([lindex $textLayout 0]/2) \
                - [dict get $config -highlightthickness]}]
        lappend textCoords [expr {[dict get $config -y] \
                + ($elementHeight * $ind)-1+[dict get $config -borderwidth] \
                - $elementHeight/2}]
        
        lappend currElInfo $textCoords

        set colors {}

        if {[dict get $config -state] == "disabled"} {
            set colors [list [dict get $config -foreground] [dict get $config -background]]
        } elseif {$isSelected < 0} {
            lappend colors [expr {[$lbox itemcget $elementInd -foreground]!={}\
                    ? [$lbox itemcget $elementInd -foreground] 
                    : [dict get $config -foreground]}]

            lappend colors [expr {[$lbox itemcget $elementInd -background]!={}\
                    ? [$lbox itemcget $elementInd -background] 
                    : [dict get $config -background]}]
        } else {
            set colors \
                    [expr {[$lbox itemcget $elementInd -selectforeground]!={} \
                    ? [$lbox itemcget $elementInd -selectforeground] 
                    : [dict get $config -selectforeground]}]
           
            if {[$lbox itemcget $elementInd -selectbackground] != {}} {
                lappend colors [$lbox itemcget $elementInd -selectbackground]
                
                lappend colors [ComputeShadowColors $lbox \
                        [dict get $config -relief] \
                        [$lbox itemcget $elementInd -selectbackground]]
            } else {
                lappend colors [dict get $config -selectbackground]
                
                lappend colors [ComputeShadowColors $lbox \
                        [dict get $config -relief] \
                        [dict get $config -selectbackground]]
            }
        }
     
        set ty [expr {int($ty + $elementHeight)}]
        set ty1 [expr {int($ty + $elementHeight)}]
        set elementCoords [list $tx $ty $tx1 $ty1]
        
        lappend currElInfo $colors
        lappend currElInfo $elementCoords 
        lappend currElInfo $textLayout
        lappend elementsInfo $currElInfo
        incr ind 
    }

    dict set config -selelements $selElements
    dict set config -elementsinfo $elementsInfo

    # determine visible edges of selection.

    set tmp 0

    if {[lindex $xView 0] == 0.0} {
        incr tmp
    }

    if {[lindex $xView 1] == 1.0} {
        incr tmp 2
    }

    # 1 means right edge is visible.
    # 2 means left ednge is visible.
    
    dict set config -visibleedges $tmp
    dict set config -elementheight $elementHeight

	return $config
}

#############################################################################
# proc GetMenubuttonConfig --
#
#       get and compute configuration values specific to menubutton widget.
#       this procedure is only wrapper procedure for getButtonConfig.
#
# arguments:
#       mb - a widget's path.
# results:
#       after this procedure is called, config dictionary will have all
#       information which is needed to export widget's representation to pdf.

proc GetMenubuttonConfig {mb} {
    
    return [GetButtonConfig $mb]
}

#############################################################################
# proc GetMenuConfig --
#
#       get and compute configuration values specific to menu widget.
#
# arguments:
#       m - a widget's path.
# results:
#       after this procedure is called, config dictionary will have all
#       information which is needed to export widget's representation to pdf.

proc GetMenuConfig {m} {
    
    set config [dict create]

    GetWidgetStdConfig $m config
    
    #dict set config -borderwidth [expr [dict get $config -borderwidth] *2 ]

    set numOfEntries [$m index end]
    set entriesInfo {}

    for {set i 1} {$i <= $numOfEntries} {incr i} {
        
        dict set entryDict -x [expr {[$m xposition $i] + [dict get $config -x]}]
        dict set entryDict -y [expr {[$m yposition $i] + [dict get $config -y]}]
        dict set entryDict -text [$m entrycget $i -label]
        dict set entryDict -type [$m type $i]
        dict set entryDict -image [$m entrycget $i -image]
        
        if {[dict get $entryDict -image] != ""} { 
            dict set entryDict -imagewidth \
                    [image width [$m entrycget $i -image]]
            dict set entryDict -imageheight \
                    [image height [$m entrycget $i -image]]
        }
                
        if {[$m entrycget $i -font] == ""} {
            dict set entryDict -font [$m cget -font]
        } else {
            dict set entryDict -font [$m entrycget $i -font]
        }

        if {[$m entrycget $i -background] == ""} {
            dict set entryDict -bg [$m cget -background]
        } else {
            dict set entryDict -bg [$m entrycget $i -background]
        }

        if {[$m entrycget $i -state] == "disabled"} {
            dict set entryDict -fg [$m cget -disabledforeground]
        } elseif {[$m entrycget $i -foreground] != ""} {
            dict set entryDict -fg [$m entrycget $i -foreground]
        } else {
            dict set entryDict -fg [$m cget -foreground]
        }
       
        if {(([$m type $i] == "checkbutton") || ([$m type $i] == "radiobutton") \
                && [$m entrycget $i -indicatoron])} {
            
            set indicator 1

            if {[$m type $i] == "radiobutton"} {
                set onValue [$m entrycget $i -value]
            } else {
                set onValue [$m entrycget $i -onvalue]
            }

            set currVal [eval set x "$\::[$m entrycget $i -variable]"]
            dict set entryDict -selected [expr {$currVal == $onValue ? 1 : 0}]
            
            dict set entryDict -selectcolor \
                    [expr {[$m entrycget $i -selectcolor] != "" \
                    ? [$m entrycget $i -selectcolor] : [$m cget -selectcolor]}]
            
            if {[$m entrycget $i -selectimage] != "" \
                    && [dict get $entryDict -selected]} {
                dict set entryDict -image [$m entrycget $i -selectimage]
            }
        } else {
            set indicator 0
        }

        set entrySize [ComputeMenuLabelSize [$m entrycget $i -label] \
                [dict get $entryDict -font] [$m entrycget $i -image] \
                [$m entrycget $i -compound]]
        
        set width [lindex $entrySize 0]
        set height [expr {[lindex $entrySize 1] \
                + 2 * [$m cget -activeborderwidth] + 10}]
       
        set sizeOfInd [ComputeMenuIndSpace [$m type $i] \
                [$m entrycget $i -hidemargin] $indicator \
                [$m entrycget $i -image] $width $height \
                [$m cget -borderwidth]]
        
        dict set entryDict -ispace [lindex $sizeOfInd 0] 
        
        incr width [expr {[lindex $sizeOfInd 0] \
                + 2 * [$m cget -activeborderwidth] + 10}]
        
        dict set entryDict -x1 [expr {[$m xposition $i] + $width \
                - 1 + [dict get $config -x]}]
        dict set entryDict -y1 [expr {[$m yposition $i] + $height \
                - 1 + [dict get $config -y]}]

        set textAndImageCoords [MenuGetTextAndImageCoords $entryDict \
                [$m cget -activeborderwidth] \
                [$m entrycget $i -compound] $height]
        
        dict set entryDict -textcoords [lindex $textAndImageCoords 0]
        dict set entryDict -imagecoords [lindex $textAndImageCoords 1]

        if {[dict get $entryDict -image] != ""} {
            foreach {coords width height name data ddd} [GetImageData \
                    [dict get $entryDict -imagecoords] [dict get $entryDict -image] \
                    "nw"] {
                dict set entryDict -imagecoords $coords
                dict set entryDict -imagename $name
                dict set entryDict -data $data
            }
        }

        set textLayout [ComputeTextLayout [dict get $entryDict -text] \
                        [dict get $entryDict -font] 0]
        dict set entryDict -textwidth [lindex $textLayout 0]
        dict set entryDict -textheight [lindex $textLayout 1]

        if {[$m entrycget $i -foreground] != ""} {
            dict set entryDict -foreground [$m entrycget $i -foreground]
        } else {
            dict set entryDict -foreground [dict get $config -foreground]
        }

        dict set entryDict -ix [expr {[$m xposition $i] \
                + [$m cget -activeborderwidth] + 2 \
                + ([dict get $entryDict -ispace] / 2) - 5}]
        
        dict set entryDict -iy [expr {[$m yposition $i] + ($height / 2) - 5}]
        if {[dict get $entryDict -type] == "checkbutton"} {
            dict set entryDict -ix1 [expr {[dict get $entryDict -ix] + 10}]
            dict set entryDict -iy1 [expr {[dict get $entryDict -iy] + 10}]
        } elseif {[dict get $entryDict -type] == "radiobutton"} {
            dict set entryDict -ix1 [expr {[dict get $entryDict -ix] + 8}]
            dict set entryDict -iy1 [expr {[dict get $entryDict -iy] + 8}]
        }

        if {[dict exists $entryDict -ix1]} {
            dict set entryDict -markcoords [ComputeMarkCoords [dict get $entryDict -ix] \
                    [dict get $entryDict -iy] [dict get $entryDict -ix1] \
                    [dict get $entryDict -iy1]]
        }
        
        lappend entriesInfo $entryDict
        unset entryDict
    }
    
    dict set config -entriesinfo $entriesInfo
    dict set config -highlightthickness 0

    return $config
}

#############################################################################
# proc GetCanvasConfig -- 
# 
#       get and compute configuration values specific to canvas. also compute
#       configuration values for canvas'es items.
#
# arguments:
#       c - a widget's path.
# results:
#       after this procedure is called, config dictionary will have all
#       information which is needed to export widget's representation to pdf.

proc GetCanvasConfig {c} {
 
    # get info about canvas widget

    set config [dict create]
    GetWidgetStdConfig $c config

    set tx [dict get $config -x]
    set ty [dict get $config -y]
   
    # get the area on canvas where objects are placed.
    
    set bbox [$c bbox all]
  
    # if there isn't any elements in canvas finish the procedure.

    if {$bbox == ""} {
        dict set config -itemsinfo {}
        
        return $config
    }
    
    foreach {bbx bby bbx1 bby1} $bbox break

    # get info about canvas'es items.

    set itemsInfo {}

    foreach id [$c find overlapping $bbx $bby $bbx1 $bby1] {

        if {[$c type $id] == "window"} {
            continue
        }
        
        dict set itemDict -class [$c type $id]
       
        set coords {} 
        
        foreach {x y} [$c coords $id] {

            set x [expr $x + $tx - [dict get $config -highlightthickness]]
            set y [expr $y + $ty - [dict get $config -highlightthickness]]
            
            lappend coords $x
            lappend coords $y
        }
        
        dict set itemDict -coords $coords
        
        CanvasGetItemOptions $c $id $coords itemDict

        lappend itemsInfo $itemDict
        
        unset itemDict
    }
  
    dict set config -itemsinfo $itemsInfo

    return $config
}

proc GetScaleConfig {sc} {

    set config [dict create]

    # get configuration values
    
    GetWidgetStdConfig $sc config
    
    set shadows [ComputeShadowColors $sc "sunken" \
            [dict get $config -background]]
    dict set config -troughdarkshadow [lindex $shadows 0]
    dict set config -troughlightshadow [lindex $shadows 1]
    
    set sliderCoords [$sc coords]
    
    set shadows [ComputeShadowColors $sc [dict get $config -sliderrelief]\
            [dict get $config -background]]
    dict set config -sliderdarkshadow [lindex $shadows 0]
    dict set config -sliderlightshadow [lindex $shadows 1]

    if {[dict get $config -orient] == "horizontal"} {
       
        set xSliderCoords [lindex $sliderCoords 0]

        dict set config -troughx [expr {[dict get $config -x] \
                + [dict get $config -borderwidth]}]
        dict set config -troughx1 [expr {[dict get $config -x1] \
                - [dict get $config -borderwidth]}]

        for {set i 0} {[$sc identify $xSliderCoords $i] != "slider"} {incr i} {}
        dict set config -troughy [expr {$i + [dict get $config -y] \
                - 1 - [dict get $config -highlightthickness]}]

        for {} {[$sc identify $xSliderCoords $i] == "slider"} {incr i} {}
        dict set config -troughy1 [expr {$i + [dict get $config -y] \
                - 2 - [dict get $config -highlightthickness]}]

        dict set config -sliderx [expr {[lindex $sliderCoords 0] \
                - [dict get $config -sliderlength]/2 \
                + [dict get $config -x] \
                - [dict get $config -highlightthickness]}]
        dict set config -sliderx1 [expr {[lindex $sliderCoords 0] \
                + [dict get $config -sliderlength]/2 \
                + [dict get $config -x] - 1 \
                - [dict get $config -highlightthickness]}]
        dict set config -slidery [expr {[dict get $config -troughy] \
                + [dict get $config -borderwidth]}]
        dict set config -slidery1 [expr {[dict get $config -troughy1] \
                - [dict get $config -borderwidth]}]

        dict set config -spartx [expr {[dict get $config -sliderx] \
                + [dict get $config -borderwidth] / 2}]
        dict set config -sparty [expr {[dict get $config -slidery] \
                + [dict get $config -borderwidth] / 2}]
        dict set config -spartx1 [expr {[lindex $sliderCoords 0] \
                + [dict get $config -x] - 2 \
                - [dict get $config -highlightthickness]}]
        dict set config -sparty1 [expr {[dict get $config -slidery1] \
                - [dict get $config -borderwidth] / 2}]
        
        dict set config -spartpx [expr {[lindex $sliderCoords 0] \
                + [dict get $config -x] - 1 \
                - [dict get $config -highlightthickness]}]
        dict set config -spartpy [dict get $config -sparty]
        dict set config -spartpx1 [expr {[dict get $config -sliderx1] \
                - [dict get $config -borderwidth] / 2}]
        dict set config -spartpy1 [dict get $config -sparty1]

        set tick {}
        if {[dict get $config -tickinterval]} {
            for {set i [dict get $config -from]} \
                    {$i < [dict get $config -to]} \
                    {set i [expr {$i + [dict get $config -tickinterval]}]} {
            
                set coords [$sc coords $i]
                set val [$sc get {*}$coords]
                set textLayout [ComputeTextLayout $val \
                        [dict get $config -font] 0]
                
                # pDF: need change of text's y coordinates
                
                set xCoords [lindex $coords 0]
                
                if {[expr {$xCoords - (([lindex $textLayout 0]/2) \
                    + 2 + [dict get $config -borderwidth]) \
                        - [dict get $config -highlightthickness]}] < 0} {
                
                    set xCoords [expr {([lindex $textLayout 0]/2) \
                            + 2 + [dict get $config -borderwidth] \
                            + [dict get $config -highlightthickness]}]
                }
             
                if {[expr {$xCoords + (([lindex $textLayout 0]/2) \
                        + 2 + [dict get $config -borderwidth] \
                        + [dict get $config -x]) \
                        - [dict get $config -highlightthickness]}] \
                        > [dict get $config -x1]} {
                
                    set xCoords [expr {[dict get $config -x1] \
                            - [dict get $config -borderwidth] \
                            - ([lindex $textLayout 0]/2) - 2}]
                } else {
                    incr xCoords [expr {[dict get $config -x] \
                            - [dict get $config -highlightthickness]}]
                }

                lset coords 0 $xCoords
                lset coords 1 [expr {([dict get $config -y1] \
                        - [dict get $config -troughy1] \
                        - [dict get $config -borderwidth]) / 2 \
                        + [dict get $config -troughy1] + 1}]
                
                lappend tick "[list $coords] $val $textLayout"
            }
        }
        
        dict set config -tick $tick
     
        # a heigh of space between slider's through and top of widget,
        # without border's width.

        set height [expr {[dict get $config -troughy] \
                - [dict get $config -y] - [dict get $config -borderwidth]}]

        if {[dict get $config -showvalue]} {
            set coords [$sc coords]
            set textLayout [ComputeTextLayout [$sc get] \
                    [dict get $config -font] 0]
           
            set xCoords [lindex $coords 0]
            
            if {[expr {$xCoords - (([lindex $textLayout 0]/2) \
                    + 2 + [dict get $config -borderwidth]) \
                        - [dict get $config -highlightthickness]}] < 0} {
                
                set xCoords [expr {([lindex $textLayout 0]/2) \
                        + 2 + [dict get $config -borderwidth] \
                        + [dict get $config -highlightthickness]}]
            }
             
            if {[expr {$xCoords + (([lindex $textLayout 0]/2) \
                    + 2 + [dict get $config -borderwidth] \
                    + [dict get $config -x])}] > [dict get $config -x1]} {
            
                set xCoords [expr {[dict get $config -x1] \
                        - [dict get $config -borderwidth] \
                        - ([lindex $textLayout 0]/2) - 2}]
            } else {
                incr xCoords [expr {[dict get $config -x] \
                        - [dict get $config -highlightthickness]}]
            }
           
            lset coords 0 $xCoords
            if {[dict get $config -label] != ""} {
                lset coords 1 [expr {(3 * $height / 4)+[dict get $config -y]\
                        + [dict get $config -borderwidth]}]
            } else {
                lset coords 1 [expr {($height / 2)+[dict get $config -y] \
                        + [dict get $config -borderwidth]}]
            }

            dict set config -valuecoords $coords
            dict set config -valuetextlayout $textLayout
            dict set config -value [$sc get]
        }
        # compute scale's label coordinates
        
        if {[dict get $config -label] != ""} {
            set textLayout [ComputeTextLayout [dict get $config -label] \
                    [dict get $config -font] 0]
            lset coords 0 [expr {[dict get $config -borderwidth] \
                    + ([lindex $textLayout 0]/2) + [dict get $config -x] \
                    + ([font metrics [dict get $config -font] -ascent]/2)}]
            if {[dict get $config -showvalue]} { 
                lset coords 1 [expr {($height / 4) + [dict get $config -y] \
                        + [dict get $config -borderwidth]}]
            } else {
                lset coords 1 [expr {($height / 2) + [dict get $config -y] \
                        + [dict get $config -borderwidth]}]
            }

            dict set config -labelcoords $coords
            dict set config -labeltextlayout $textLayout
        }
    } else {
       
        # a coordinates of center of slider.

        set ySliderCoords [lindex $sliderCoords 1]

        dict set config -troughy [expr {[dict get $config -y] \
                + [dict get $config -borderwidth]}]
        dict set config -troughy1 [expr {[dict get $config -y1] \
                - [dict get $config -borderwidth]}]

        for {set i 0} {[$sc identify $i $ySliderCoords] != "slider"} {incr i} {}
        dict set config -troughx [expr {$i + [dict get $config -x] \
                - 1 - [dict get $config -highlightthickness]}]

        for {} {[$sc identify $i $ySliderCoords] == "slider"} {incr i} {}
        dict set config -troughx1 [expr {$i + [dict get $config -x] \
                - 2 - [dict get $config -highlightthickness]}]

        dict set config -sliderx [expr {[dict get $config -troughx] \
                + [dict get $config -borderwidth]}]
        dict set config -sliderx1 [expr {[dict get $config -troughx1] \
                - [dict get $config -borderwidth]}]

        dict set config -slidery [expr {[lindex $sliderCoords 1] \
                - [dict get $config -sliderlength]/2 \
                + [dict get $config -y] \
                - [dict get $config -highlightthickness]}]
        dict set config -slidery1 [expr {[lindex $sliderCoords 1] \
                + [dict get $config -sliderlength]/2 \
                + [dict get $config -y] - 1 \
                - [dict get $config -highlightthickness]}]

        dict set config -spartx [expr {[dict get $config -sliderx] \
                + [dict get $config -borderwidth] / 2}]
        dict set config -sparty [expr {[dict get $config -slidery] \
                + [dict get $config -borderwidth] / 2}]
        dict set config -spartx1 [expr {[dict get $config -sliderx1] \
                - [dict get $config -borderwidth] / 2}]
        dict set config -sparty1 [expr {[lindex $sliderCoords 1] \
                + [dict get $config -y] - 1 \
                - [dict get $config -highlightthickness]}]
        
        dict set config -spartpx [dict get $config -spartx]
        dict set config -spartpy [expr {[dict get $config -sparty1] + 1}]
        dict set config -spartpy1 [expr {[dict get $config -slidery1] \
                - [dict get $config -borderwidth] / 2}]
        dict set config -spartpx1 [dict get $config -spartx1]

        set tick {}
        if {[dict get $config -tickinterval]} {
            for {set i [dict get $config -from]} \
                    {$i < [dict get $config -to]} \
                    {set i [expr {$i + [dict get $config -tickinterval]}]} {
            
                set coords [$sc coords $i]
                set val [$sc get {*}$coords]
                set textLayout [ComputeTextLayout $val \
                        [dict get $config -font] 0]
                
                lset coords 0 [expr {[dict get $config -x] \
                        + ([lindex $textLayout 0]/2) + 2 \
                        + [font metrics [dict get $config -font] -ascent]/2}]
                lset coords 1 [expr {[dict get $config -y] \
                        + [lindex $coords 1] \
                        - [dict get $config -highlightthickness]}]
                
                lappend tick "[list $coords] $val $textLayout"
            }
        }
        
        dict set config -tick $tick
     
        # a heigh of space between slider's through and top of widget,
        # without border's width.

        set height [expr {[dict get $config -troughy] \
                - [dict get $config -y] - [dict get $config -borderwidth]}]

        if {[dict get $config -showvalue]} {
            set coords [$sc coords]
            set textLayout [ComputeTextLayout [$sc get] \
                    [dict get $config -font] 0]
           
            set yCoords [lindex $coords 1]
            
            lset coords 0 [expr {[dict get $config -troughx] \
                    - ([lindex $textLayout 0]/2) - 2}]
            
            lset coords 1 [expr {$yCoords + [dict get $config -y] \
                    - [dict get $config -highlightthickness]}]

            dict set config -valuecoords $coords
            dict set config -valuetextlayout $textLayout
            dict set config -value [$sc get]
        }
        
        # compute scale's label coordinates
        
        if {[dict get $config -label] != ""} {
            set textLayout [ComputeTextLayout [dict get $config -label] \
                    [dict get $config -font] 0]
            set asc [expr {[font metrics [dict get $config -font] -ascent]/2}]
            
            lset coords 0 [expr {[dict get $config -borderwidth] \
                    + ([lindex $textLayout 0]/2) + [dict get $config -sliderx1] \
                    + $asc}]

            lset coords 1 [expr {$asc + [dict get $config -y] \
                    + [dict get $config -borderwidth] \
                    + ([lindex $textLayout 1]/2)}]
            
            dict set config -labelcoords $coords
            dict set config -labeltextlayout $textLayout
        }
    } 
    
    return $config
}

proc GetScrollbarConfig {sc} {

    set config [dict create]
    GetWidgetStdConfig $sc config

    # set color for slider and arrows

    set shadows [ComputeShadowColors $sc "sunken" \
            [dict get $config -background]]
    dict set config -elementsshadows $shadows

    if {[dict get $config -orient] == "horizontal"} {
        # get coordinates of arrows and slider.
        
        set height [expr {([dict get $config -height] \
                + [dict get $config -highlightthickness]) / 2}]
        set xPos 0
        
        while {[$sc identify $xPos $height] != "arrow1"} {
            incr xPos
        }
    
        set coords {}

        lappend coords [expr {$xPos + [dict get $config -x] \
                - [dict get $config -highlightthickness]}]
        lappend coords [expr {[dict get $config -y] + $height \
                - [dict get $config -highlightthickness]/2}] 

        while {[$sc identify $xPos $height] == "arrow1"} {
            incr xPos
        }

        lappend coords [expr {$xPos + [dict get $config -x] \
                - 2 - [dict get $config -highlightthickness]}]
        lappend coords [expr {[dict get $config -y] \
                + [dict get $config -borderwidth]}]
        
        lappend coords [lindex $coords 2]
        lappend coords [expr {[dict get $config -y1] \
                - [dict get $config -borderwidth]}]

        set interiorCoords {}
        lappend interiorCoords [expr {[lindex $coords 0] \
                + 2 * [dict get $config -elementborderwidth]}]
        lappend interiorCoords [lindex $coords 1]
        lappend interiorCoords [expr {[lindex $coords 2] \
                - [dict get $config -elementborderwidth]}]
        lappend interiorCoords [expr {[lindex $coords 3] \
                + 2*[dict get $config -elementborderwidth]}]
        lappend interiorCoords [expr {[lindex $coords 4] \
                - [dict get $config -elementborderwidth]}]
        lappend interiorCoords [expr {[lindex $coords 5] \
                - 2*[dict get $config -elementborderwidth]}]
    
        dict set config -arrowcoords $coords
        dict set config -interiorcoords $interiorCoords

        set coords {}
        lappend coords [expr {[dict get $config -x1] \
                - [dict get $config -borderwidth]}]
        lappend coords [lindex [dict get $config -arrowcoords] 1]
        
        lappend coords [expr {[dict get $config -x1] \
                - [dict get $config -borderwidth] \
                - ([lindex [dict get $config -arrowcoords] 2] \
                - [lindex [dict get $config -arrowcoords] 0])}]
        lappend coords [lindex [dict get $config -arrowcoords] 5]
        lappend coords [lindex $coords 2] 
        lappend coords [lindex [dict get $config -arrowcoords] 3]
        
        set interiorCoords {}
 
        lappend interiorCoords [expr {[lindex $coords 0] \
                - 2 * [dict get $config -elementborderwidth]}]
        lappend interiorCoords [lindex $coords 1]
        lappend interiorCoords [expr {[lindex $coords 2] \
                + [dict get $config -elementborderwidth]}]
        lappend interiorCoords [expr {[lindex $coords 3] \
                - 2*[dict get $config -elementborderwidth]}]
        lappend interiorCoords [expr {[lindex $coords 4] \
                + [dict get $config -elementborderwidth]}]
        lappend interiorCoords [expr {[lindex $coords 5] \
                + 2*[dict get $config -elementborderwidth]}]
        
        dict set config -arrowcoords1 $coords
        dict set config -interiorcoords1 $interiorCoords
        
        while {[$sc identify $xPos $height] != "slider"} {
            incr xPos
        }

        dict set config -sliderx [expr {$xPos + [dict get $config -x] \
                - [dict get $config -highlightthickness]}]
        dict set config -slidery [lindex $coords 5]
        
        while {[$sc identify $xPos $height] == "slider"} {
            incr xPos
        }

        dict set config -sliderx1 [expr {$xPos + [dict get $config -x] \
                - 2 - [dict get $config -highlightthickness]}]
        dict set config -slidery1 [lindex $coords 3]
 
        set deltaSliderX [expr {[dict get $config -sliderx1] \
                - [dict get $config -sliderx]}] 

        if {$deltaSliderX < [expr {2*[dict get $config -elementborderwidth]}]} {
                dict set config -sliderborderwidth [expr {$deltaSliderX / 2}]

                if {[dict get $config -sliderborderwidth] < 2} {
                    dict set config -sliderborderwidth 2
                }
        } else {
            dict set config -sliderborderwidth \
                    [dict get $config -elementborderwidth]
        }
    } else {
        # get coordinates of arrows and slider.
        
        set width [expr {([dict get $config -width] \
                + [dict get $config -borderwidth]) / 2}]
        
        set yPos 0
        while {[$sc identify $width $yPos] != "arrow1"} {
            incr yPos
        }

        set coords {}
        lappend coords [expr {[dict get $config -x] + $width}]
        lappend coords [expr {[dict get $config -y] + $yPos \
                - [dict get $config -highlightthickness]}] 

        while {[$sc identify $width $yPos] == "arrow1"} {
            incr yPos
        }

        lappend coords [expr {[dict get $config -x] \
                + [dict get $config -borderwidth]}]
        lappend coords [expr {$yPos + [dict get $config -y] \
                - [dict get $config -highlightthickness]}]
        
        lappend coords [expr {[dict get $config -x1] \
                - [dict get $config -borderwidth]}]
        lappend coords [lindex $coords 3]
        
        set interiorCoords {}
        lappend interiorCoords [lindex $coords 0]
        lappend interiorCoords [expr {[lindex $coords 1] \
                + 2 * [dict get $config -elementborderwidth]}]
         
        lappend interiorCoords [expr {[lindex $coords 2] \
                + 2*[dict get $config -elementborderwidth]}]
        lappend interiorCoords [expr {[lindex $coords 3] \
                - [dict get $config -elementborderwidth]}]

        lappend interiorCoords [expr {[lindex $coords 4] \
                - 2*[dict get $config -elementborderwidth]}]
        lappend interiorCoords [lindex $interiorCoords 3]
    
        dict set config -arrowcoords $coords
        dict set config -interiorcoords $interiorCoords

        set coords {}
        lappend coords [lindex [dict get $config -arrowcoords] 0]
        lappend coords [expr {[dict get $config -y1] \
                - [dict get $config -borderwidth]}]

        lappend coords [lindex [dict get $config -arrowcoords] 4]
        lappend coords [expr {[dict get $config -y1] \
                - [dict get $config -borderwidth] \
                - ([lindex [dict get $config -arrowcoords] 3] \
                - [lindex [dict get $config -arrowcoords] 1])}]
        
        lappend coords [lindex [dict get $config -arrowcoords] 2]
        lappend coords [lindex $coords 3]

        set interiorCoords {}
 
        lappend interiorCoords [lindex $coords 0]
        lappend interiorCoords [expr {[lindex $coords 1] \
                - 2 * [dict get $config -elementborderwidth]}]
         
        lappend interiorCoords [expr {[lindex $coords 2] \
                - 2*[dict get $config -elementborderwidth]}]
        lappend interiorCoords [expr {[lindex $coords 3] \
                + [dict get $config -elementborderwidth]}]

        lappend interiorCoords [expr {[lindex $coords 4] \
                + 2*[dict get $config -elementborderwidth]}]
        lappend interiorCoords [lindex $interiorCoords 3]
        
        dict set config -arrowcoords1 $coords
        dict set config -interiorcoords1 $interiorCoords
        
        while {[$sc identify $width $yPos] != "slider"} {
            incr yPos
        }

        dict set config -sliderx [expr {[dict get $config -x] \
                + [dict get $config -borderwidth]}]
        dict set config -slidery [expr {[dict get $config -y] + $yPos \
                - [dict get $config -highlightthickness]}]
        
        while {[$sc identify $width $yPos] == "slider"} {
            incr yPos
        }

        dict set config -sliderx1 [expr {[dict get $config -x1] \
                - [dict get $config -borderwidth]}]
        dict set config -slidery1 [expr {[dict get $config -y] + $yPos \
                - [dict get $config -highlightthickness]}]
 
        set deltaSliderY [expr {[dict get $config -slidery1] \
                - [dict get $config -slidery]}] 

        if {$deltaSliderY < [expr {2*[dict get $config -elementborderwidth]}]} {
                dict set config -sliderborderwidth [expr {$deltaSliderY / 2}]

                if {[dict get $config -sliderborderwidth] < 2} {
                    dict set config -sliderborderwidth 2
                }
        } else {
            dict set config -sliderborderwidth \
                    [dict get $config -elementborderwidth]
        }
    }

    return $config
}

proc GetTextConfig {txt} {

    set config [dict create]
    GetWidgetStdConfig $txt config

    return $config
}

#############################################################################
# procedure: CanvasGetItemOptions --
#
#       get item's configuration values. get only needed options.
#
# arguments:
#       c            - canvas path.
#       id           - id of element in canvas.
#       coords       - a coordinates of item.
#       itemDictName - a dictionary to be filled with configuration values
# results:
#       itemDictName dictionary is filled with configuration values.

proc CanvasGetItemOptions {c id coords itemDictName} {
    
    upvar $itemDictName itemDict

    if {[$c type $id] == "image"} {
        foreach {coords width height name data ddd} [GetImageData \
                $coords [$c itemcget $id -image] \
                [$c itemcget $id -anchor]] {
            
            dict set itemDict -coords $coords
            dict set itemDict -width $width
            dict set itemDict -height $height
            dict set itemDict -imagename $name
            dict set itemDict -data $data
            dict set itemDict -ddd $ddd
        }
        
        return
    }

    if {[$c type $id] == "text"} {
        
        set textLayout [ComputeTextLayout [$c itemcget $id -text] \
                [$c itemcget $id -font] 0]
        dict set itemDict -textwidth [lindex $textLayout 0]
        dict set itemDict -textheight [lindex $textLayout 1]
    }
    
    # options which may be in one of three states:
    # disabled, active or normal.

    dict set itemDict -fill [$c itemcget $id -fill]
    
    set options [list "-dash" "-fill" "-outline" \
            "-outlinestipple" "-stipple" "-width"]

    if {[$c itemcget $id -state] == "disabled"} {
        set state "disabled"
    } else {
        set state ""
    }

    foreach item [$c itemconfigure $id] {
        
        set optName [lindex $item 0]

        # avoid options without value.

        #if {[string length [lindex $item 4]] == 0} {
        #    continue
        #}

        if {[string match "-active*" $optName] \
                || [string match "-disabled*" $optName]||$optName == "-tags" \
                || $optName == "-state"} {
            continue
        }

        if {[set optInd [lsearch -exact $options $optName]] > 0} {
            set opt [lindex $options $optInd]
            
            if {$state == "disabled"} {
                dict set itemDict $opt [$c itemcget $id "disabled$opt"]
            } else {
                dict set itemDict $opt [$c itemcget $id $opt] 
            }
        } else {
            dict set itemDict $optName [lindex $item 4]
        }
    }

}

proc GetImageData {coords img anchor} {

    set fileName [$img cget -file]
    
    set width [$img cget -width] 
    set height [$img cget -height]
    set name $img 

    if {$anchor != "center"} {
        
        if {[regexp "n" $anchor]} {
            lset coords 1 [expr {[lindex $coords 1] + ($height / 2)}]
        }
        if {[regexp "s" $anchor]} {
            lset coords 1 [expr {[lindex $coords 1] - ($height / 2)}]
        }
        if {[regexp "e" $anchor]} {
            lset coords 0 [expr {[lindex $coords 0] - ($width / 2)}]
        }
        if {[regexp "w" $anchor]} {
            lset coords 0 [expr {[lindex $coords 0] + ($width / 2)}]
        }
    }

    if {[$img cget -data] != ""} {
        set data [$img data]
    } else {
        set fd [open $fileName "r"]
        set data [read $fd]
        set data [::base64::encode $data]
        set img [image create photo -data $data]
        set ddd $data 
       
        set data [$img data -from 0 0 $width $height]
        close $fd
    }
    
    return [list $coords $width $height $name $data $ddd]
}

#############################################################################
# procedure: MenuGetTextAndImageCoords --
#
#       get text's and image's coords in menu widget.
#
# arguments:
#       entryDict       - dictionary with entry configuration values.
#       activeBorder    - activeborderwidth value.
#       compound        - compound value.
#       h               - a height of entry.
# results:
#       a image and text coords of menu's entry are returned.

proc MenuGetTextAndImageCoords {entryDict activeBorder compound h} {

    array set c [dict get $entryDict]
    
    set textXOffset $c(-x)
    set textYOffset $c(-y)
    set imageXOffset $c(-x)
    set imageYOffset $c(-y)

    set entryWidth [expr {$c(-x1) - $c(-x)}]
    set entryHeight [expr {$c(-y1) - $c(-y)}]

    if {($c(-text) != "") && !(($c(-image) != "") && ($compound == "none"))} {
        set textLayout [ComputeTextLayout $c(-text) $c(-font) 0]
        set textWidth [lindex $textLayout 0]
        set textHeight [lindex $textLayout 1]
        
        incr textXOffset [expr {$c(-ispace) + $activeBorder + 5}]
        # pDF: delete next line.
        incr textXOffset [expr {$textWidth / 2}]
        incr textYOffset [expr {$entryHeight / 2}]
    }

    if {$c(-image) != ""} {
        set imageWidth [image width $c(-image)]
        set imageHeight [image height $c(-image)]
        
        incr imageXOffset [expr {$c(-ispace) + $activeBorder + 5}]
        # xXX i don't know why it is working.
        incr h 6
        incr imageYOffset [expr {($h - $imageHeight) / 2}]
    }

    if {($c(-text) != "") && ($c(-image) != "")} {
        set fWidth [expr {$imageWidth > $textWidth ? $imageWidth : $textWidth}]
        
        switch $compound {
            "top" {
                incr textXOffset [expr {($fWidth - $textWidth) / 2}]
                incr textYOffset [expr {($imageHeight / 2) + 2}]
                incr imageXOffset [expr {($fWidth - $imageWidth) / 2}]
                incr imageYOffset [expr {-$textHeight / 2 + 1}]
            } 
            "bottom" {
                incr textXOffset [expr {($fWidth - $textWidth) / 2}]
                incr textYOffset [expr {(-$imageHeight / 2)}]
                incr imageXOffset [expr {($fWidth - $imageWidth) / 2}]
                incr imageYOffset [expr {($textHeight / 2) + 2}]
            }
            "left" {
                incr textXOffset [expr {$imageWidth + 2}]
                incr imageXOffset -$c(-ispace)
                
                if {($c(-type)!="radiobutton") && ($c(-type)!="checkbutton")} {
                    incr textXOffset -$c(-ispace)
                }
            }
            "right" {
                incr imageXOffset [expr {$textWidth + 2}]
            }
            "center" {
                incr textXOffset [expr {($fWidth - $textWidth) / 2}]
                incr imageXOffset [expr {($fWidth - $imageWidth) / 2}]
            }
        }
    }

    return [list [list $textXOffset $textYOffset] \
            [list $imageXOffset $imageYOffset]]
}

#############################################################################
# procedure LabelframeGetLabelInfo --
# 
#       get info about label in labelframe widget.
#
# arguments:
#       config  - a dictionary with configuration values.
# results:
#       borderX - x displacement of labelframe's left border caused by label.
#       borderY - y displacement of labelframe's top border caused by label.
#       borderX1- x displacement of labelframe's right border caused by label.
#       borderY1- y displacement of labelframe's bottom border caused by label.
#       labelTextCoords - a coordinates of labelframe's text (center of text). 
#       labelBoxCoords - a coordinates of rectangular text background. 

proc LabelframeGetLabelInfo {config} {

    set labelSpace 1
    set labelMargin 2

    # how to label affects the border

    set borderX 0
    set borderX1 0
    set borderY 0
    set borderY1 0

    set labelTextCoords {0 0} 
    set labelBoxCoords {0 0 0 0}

    array set c [dict get $config]
 
    if {$c(-class) == "Labelframe"} {
        set c(-wraplength) 0
    }

    # the widget is using text instead of external label widget.

    if {$c(-labelwidget) == ""} {
        set textLayout [ComputeTextLayout $c(-text) $c(-font) $c(-wraplength)]

        set labelReqWidth [expr {[lindex $textLayout 0] + 2 * $labelSpace}]
        set labelReqHeight [expr {[lindex $textLayout 1] + 2 * $labelSpace}]
        
        # make sure label size is at least as big as the border.

        if {[string match {*[sn]*} $c(-labelanchor)]} {
            if {$labelReqHeight < $c(-borderwidth)} {
                set labelReqHeight $c(-borderwidth)
            }
        } else {
            if {$labelReqWidth < $c(-borderwidth)} {
                set labelReqWidth $c(-borderwidth)
            }
        }

        set padding 0 ;#$c(-highlightthickness)
        if {$c(-borderwidth) > 0} {
            incr padding [expr {$c(-borderwidth) + $labelMargin}]
        }

        incr padding $padding

        set maxWidth $c(-width)
        set maxHeight $c(-height)

        if {[string match {*[sn]*} $c(-labelanchor)]} {
            incr maxWidth -$padding
            if {$maxWidth < 1} {
                set maxWidth 1
            }
        } else {
            incr maxHeight -$padding
            if {$maxHeight < 1} {
                set maxHeight 1
            }
        }

        set labelWidth $labelReqWidth
        set labelHeight $labelReqHeight
        
        if {$labelWidth > $maxWidth} {
            set labelWidth $maxWidth
        }
        if {$labelHeight > $maxHeight} {
            set labelHeight $maxHeight
        }

        set otherWidth [expr {$c(-width) - $labelWidth}]
        set otherHeight [expr {$c(-height) - $labelHeight}]
        set otherWidthT [expr {$c(-width) - $labelReqWidth}]
        set otherHeightT [expr {$c(-height) - $labelReqHeight}]

        switch -glob $c(-labelanchor) {
            "n*" {
                lset labelTextCoords 1 0
                lset labelBoxCoords 1 0
            }
            "s*" {
                lset labelTextCoords 1 $otherHeightT
                lset labelBoxCoords 1 $otherHeight
            }
            "e*" {
                lset labelTextCoords 0 $otherWidthT
                lset labelBoxCoords 0 $otherWidth
            }
            default {
                lset labelTextCoords 0 0
                lset labelBoxCoords 0 0
            }
        }

        if {$c(-borderwidth) > 0} {
            # XXX: Delete 0*$c(-highli...) if sure this is not needed. 
            set padding [expr {$c(-borderwidth) + $labelMargin + \
                    0*$c(-highlightthickness)}]
        }

        switch -glob $c(-labelanchor) {
            "?w" {
                lset labelTextCoords  0 $padding
                lset labelBoxCoords 0 $padding
            }
            "[ns]" {
                lset labelTextCoords 0 [expr {$otherWidthT / 2}]
                lset labelBoxCoords 0 [expr {$otherWidth / 2}]
            }
            "?e" {
                lset labelTextCoords 0 [expr {$otherWidthT - $padding}]
                lset labelBoxCoords 0 [expr {$otherWidth - $padding}]
            }
            "?n" {
                lset labelTextCoords 1 $padding
                lset labelBoxCoords 1 $padding
            }
            "[ew]" {
                lset labelTextCoords 1 [expr {$otherHeightT/2}]
                lset labelBoxCoords 1 [expr {$otherHeight/2}]
            }
            default {
                lset labelTextCoords 1 [expr {$otherHeightT - $padding}]
                lset labelBoxCoords 1 [expr {$otherHeight - $padding}]
            }
        }

        lset labelBoxCoords 0 [expr {[lindex $labelBoxCoords 0] + $c(-x)}]
        lset labelBoxCoords 1 [expr {[lindex $labelBoxCoords 1] + $c(-y)}]
        lset labelBoxCoords 2 [expr {[lindex $labelBoxCoords 0]+$labelWidth-1}]
        lset labelBoxCoords 3 [expr {[lindex $labelBoxCoords 1]+$labelHeight-1}]

        lset labelTextCoords 0 [expr {[lindex $labelTextCoords 0] + $c(-x)}]
        lset labelTextCoords 1 [expr {[lindex $labelTextCoords 1] + $c(-y)}]

        # pDF: the next two lines will be deleted when generating pdf output
        lset labelTextCoords 0 [expr {[lindex $labelTextCoords 0] + $labelWidth/2 }]
        lset labelTextCoords 1 [expr {[lindex $labelTextCoords 1] + $labelHeight/2}]

    } else {
        # the widget is using external label as it's label.
       
        set labelWidget $c(-labelwidget)
        set labelWidth [winfo width $labelWidget]
        set labelHeight [winfo height $labelWidget]
    }

    switch -glob $c(-labelanchor) {
        "n*" {
            set borderY [expr {($labelHeight - $c(-borderwidth) + 1) / 2}] 
        }
        "s*" {
            set borderY1 [expr {($labelHeight - $c(-borderwidth)) / 2}]
        }
        "w*" {
            set borderX [expr {($labelWidth - $c(-borderwidth)) / 2}]
        }
        default {
            set borderX1 [expr {($labelWidth - $c(-borderwidth)) / 2}]
        }
    }

    return [list $borderX $borderY $borderX1 $borderY1 \
            $labelTextCoords $labelBoxCoords]
}

# procedure: btnGetTextLength - get length of text
proc BtnGetTextLength {config} {

    return [lindex [ComputeTextLayout $config] 0]
}

# procedure: btnGetTextHeight - get height of text in button 
proc BtnGetTextHeight {config} {

    return [lindex [ComputeTextLayout $config] 1]
}

proc ComputeTextLayout {text font wraplength} {

    set height 0
    set width 0

    # if wraplength option is enabled we also return new string
    # which contains '\n' at points where text was wrapped
    set newText ""

    if {$text == ""} {
        return [list 0 0]
    }

    set lineSpace [font metrics $font -linespace]

    # get index of \n in string and interpret '\r' special character	
    set sList {}

    foreach ch {"\r" "\n"} {
        set tmp 0
        while {1} {
            set tmp [string first $ch $text $tmp]

            if {$tmp == -1} {
                break
            }

            if {$ch == "\r"} {
                incr tmp
                set text [string range $text $tmp end]
            } else {
                lappend sList $tmp
                incr tmp
            }
        }
    }

    if {!$wraplength} {
        set height [expr {([llength $sList] + 1) * $lineSpace}]
        
        # choose the longest piece of text
        set start 0
        set i 0
        lappend sList [string length $text]

        foreach chunk $sList {
            set tmp [string range $text $start [expr [lindex $sList $i]-1]]
            set len [font measure $font  $tmp]
            
            if {$len > $width} {
                set width $len
            }
            
            set start [expr {[lindex $sList $i] + 1}]
            incr i
        }
    } else {

        set textLength [string length $text]
        set actualWidth 0
        set tmpString ""
        set lastSpaceIndex 0
        set lastChunkStart 0
        set char 0
        set longestChunk 0

        for {set i 0} {$i < $textLength} {incr i} {
            set char [string index $text $i]

            if {$char == " "} {
                set lastSpaceIndex $i
            }
            
            if {$char == "\t"} {
                set char "        "
            }
 
            if {$char == "\n"} {
                set tmp [string range $text $lastChunkStart [expr {$i-1}]]
                
                set tmpWidth [font measure $font $tmp]
                if {$tmpWidth > $width} { 
                    set width $tmpWidth
                }

                append newText "$tmp\n"

                incr height
                set lastChunkStart [expr {$i + 1}]
                set lastSpaceIndex 0
                set tmpString ""

                continue
            }

            if {[font measure $font "$tmpString$char"] > $wraplength} {
                
                if {$lastSpaceIndex != 0} {
                    set tmp [string range $text \
                            $lastChunkStart $lastSpaceIndex]
                    set tmpWidth [font measure $font $tmp]
                    
                    if {$tmpWidth > $width} {
                        set width $tmpWidth
                    }

                    #incr height
                    #append newText "\n"
                    set i $lastSpaceIndex
                    set lastSpaceIndex 0
                    #set lastChunkStart $i
                    #set tmpString ""
                } else {
                    set tmp [string range $text \
                            $lastChunkStart [expr {$i-1}]]
                
                    set tmpWidth [font measure $font $tmp]
                    if {$tmpWidth > $width} { 
                        set width $tmpWidth
                    }
                    
                    incr height
                    append newText "$tmp\n"

                    set lastChunkStart $i
                    set tmpString ""

                    if {$char == "        "} {
                        set char ""
                        set text [string replace $text $i $i ""]
                    }
                }
            }
            
            append tmpString $char
        }
        
        if {$tmpString != ""} {
            set tmpWidth [font measure $font $tmpString]
            if {$tmpWidth > $width} {
                set width $tmpWidth
            }
            
            incr height
            append newText $tmpString
        }

        set height [expr {$height * $lineSpace}]
        #set width [font measure $c(-font) "This is a"]

    }
   
    if {$wraplength != 0} {
        return [list $width $height $newText]
    } else {
        return [list $width $height]
    }
}

proc BtnGetTextCoords {config innerx innery dx dy} {

    array set c [dict get $config]

    # compute the displacement caused by anchor
    set anchor [ComputeAnchor $config $innerx $innery $c(-padx) $c(-pady)]
    set textXOffset [expr {[lindex $anchor 0] + $c(-x) + $dx}]
    set textYOffset [expr {[lindex $anchor 1] + $c(-y) + $dy}]
    
    # we must repair text coords because anchor returns
    # the top left coords, when we generate pdf output
    # this is not needed
    # pDF: delete the next two lines when generate pdf output

    set textXOffset [expr {$textXOffset + ($c(-textwidth)/2)}]
    set textYOffset [expr {$textYOffset + ($c(-textheight)/2)}]

    # if this is a button widget we must shligtly change it's coordinates
    if {$c(-class) == "Button"} {
        set textXOffset [expr {$textXOffset + 1}]
        set textYOffset [expr {$textYOffset + 1}]
        if {$c(-relief) == "sunken"} {
            set textXOffset [expr {$textXOffset + 1}]
            set textYOffset [expr {$textYOffset + 1}]
        } elseif {$c(-relief) == "raised"} {
            set textXOffset [expr {$textXOffset - 1}]
            set textYOffset [expr {$textYOffset - 1}]
        }
    }

    if {$c(-class) == "Radiobutton" || $c(-class) == "Checkbutton"} {
        incr textXOffset $c(-ispace)
    }
   
    return [list $textXOffset $textYOffset]
}

proc BtnGetImageCoords {config innerx innery padx pady dx dy} {

    array set c [dict get $config]
   
    # compute the displacement caused by anchor
    
    set anchor [ComputeAnchor $config $innerx $innery $padx $pady]

    set imageXOffset [expr {[lindex $anchor 0] + $c(-x) + $dx}]
    set imageYOffset [expr {[lindex $anchor 1] + $c(-y) + $dy}]
    
    # if this is a button widget we must shligtly change it's coordinates
   
    if {$c(-class) == "Button"} {

        incr imageXOffset
        incr imageYOffset
        if {$c(-relief) == "sunken"} {
                incr imageXOffset
                incr imageYOffset
        } elseif {$c(-relief) == "raised"} {
                incr imageXOffset -1
                incr imageYOffset -1
        }
    }

    incr imageXOffset [expr {$c(-ispace)}]

    return [list $imageXOffset $imageYOffset]
}

proc BtnGetTextAndImageCoords {config} {

    # full width and full height

    set fWidth 0
    set fHeight 0
  
    # text and image offsets

    set tXOffset 0
    set tYOffset 0
    set iXOffset 0
    set iYOffset 0

    # rewrite dictionary to array to short notation
   
    array set c [dict get $config]

    switch $c(-compound) {
        "top" - "bottom" {
            if {$c(-compound) == "top"} {
                set tYOffset [expr {$c(-imageheight)+$c(-pady)}]
            } else {
                set iYOffset [expr {$c(-textheight)+$c(-pady)}]
            }

            set fHeight [expr {$c(-imageheight) + $c(-textheight) + $c(-pady)}]
            set fWidth [expr {$c(-imagewidth) > $c(-textwidth) \
                ? $c(-imagewidth) : $c(-textwidth)}]
            set tXOffset [expr {($fWidth - $c(-textwidth))/2}]
            set iXOffset [expr {($fWidth - $c(-imagewidth))/2}]
        }
        "left" - "right" {
            if {$c(-compound) == "left"} {
                set tXOffset [expr {$c(-imagewidth) + $c(-padx)}]
            } else {
                set iXOffset [expr {$c(-textwidth) + $c(-padx)}]
            }

            set fWidth [expr {$c(-textwidth) + $c(-padx) + $c(-imagewidth)}]
            set fHeight [expr {$c(-imageheight) > $c(-textheight) \
                    ? $c(-imageheight) : $c(-textheight)}]
            set tYOffset [expr {($fHeight - $c(-textheight))/2}]
            set iYOffset [expr {($fHeight - $c(-imageheight))/2}]
        }
        "center" {
            set fWidth [expr {$c(-imagewidth) > $c(-textwidth) \
                    ? $c(-imagewidth) : $c(-textwidth)}]
            set fHeight [expr {$c(-imageheight) > $c(-textheight) \
                    ? $c(-imageheight) : $c(-textheight)}]
            set tXOffset [expr {($fWidth - $c(-textwidth))/2}]
            set iXOffset [expr {($fWidth - $c(-imagewidth))/2}]
            set tYOffset [expr {($fHeight - $c(-textheight))/2}]
            set iYOffset [expr {($fHeight - $c(-imageheight))/2}]
        }
    }

    set textcoords [BtnGetTextCoords $config [expr {$fWidth+$c(-ispace)}] \
        $fHeight $tXOffset $tYOffset]

    set imageCoords [BtnGetImageCoords $config \
        [expr {$fWidth+$c(-ispace)}] $fHeight \
        $c(-padx) $c(-pady) $iXOffset $iYOffset]

    return [list $textcoords $imageCoords]
}

# check whether checkbutton/radiobutton is checked
proc ButtonIsSelected {widget} {

    set class [winfo class $widget]
    
    if {$class != "Checkbutton" && $class != "Radiobutton"} {
        return 0
    }

    # get a value from wiget's variable
    set tmp [eval set x "$\::[$widget cget -variable]"]

    if {($class == "Checkbutton" && $tmp == [$widget cget -onvalue]) \
            || ($class == "Radiobutton" && $tmp == [$widget cget -value])} {
        return 1
    }

    return 0
}

proc MenuButtonIsSelected {type var onvalue} {

    if {($type != "checkbutton") && ($type != "radiobutton")} {
        return 0
    }

    #set val [eval set x "$\::"]
}

proc HasFocus {window} {

    if {[focus -displayof $window] == $window} {
        return 1
    }

    return 0
}

# compute shadows for button's border
proc ComputeShadowColors {window relief background} {

    # if the relief option is solid, we don't need
    # to compute shadow colors
    
    if {$relief == "solid"} {
        return [list black black]
    }	
   
    # if we process labelframe widget it is possible that there is no 
    # background color, end procedure in this case.

    if {$background == ""} {
        return [list black black]
    }

    # the maximal intensity
   
    set maxCol 65535

    set rgb [winfo rgb $window $background]
    set r [lindex $rgb 0]
    set g [lindex $rgb 1]
    set b [lindex $rgb 2]
    set dark #
    set light #

    foreach c $rgb {
        append dark [format "%04x" [expr {60 * $c / 100}]]
    }
    
    # compute light color
    if {$g > [expr {$maxCol * 0.95}]} {
        foreach c $rgb {
            append light [format "%04x" [expr {90 * $c / 100}]]
        }
    } else {
        set tmp1 [expr {(14 * $r) / 10}]
        if {$tmp1 > $maxCol} {
            set tmp1 $maxCol
        }
        set tmp2 [expr {($maxCol + $r) / 2}]
        append light [format "%x" [expr {$tmp1 > $tmp2 ? $tmp1:$tmp2}]]

        set tmp1 [expr {(14 * $g) / 10}]	
        if {$tmp1 > $maxCol} {
            set tmp1 $maxCol
        }
        set tmp2 [expr {($maxCol + $g) / 2}]	
        append light [format "%x" [expr {$tmp1 > $tmp2 ? $tmp1:$tmp2}]]

        set tmp1 [expr {(14 * $b) / 10}]
        if {$tmp1 > $maxCol} {
            set tmp1 $maxCol
        }
        set tmp2 [expr {($maxCol + $b) / 2}]	
        append light [format "%x" [expr {$tmp1 > $tmp2 ? $tmp1:$tmp2}]]
    }

    # if relief == sunken/groove, the shadow colors are reversed
    if {($relief == "sunken")
            || ($relief == "groove")} {
        return [list $light $dark]
    } else {
        return [list $dark $light]
    }
}

# compute indicator's space and diameter
proc ComputeIndicatorSpace {config} {

    array set c [dict get $config]
    
    if {($c(-class) != "Radiobutton") && ($c(-class) != "Checkbutton")} {
        return [list 0 0]
    }

    if {!$c(-indicatoron)} {
        return [list 0 0]
    }

    if {$c(-haveimage) && $c(-havetext) && $c(-compound) != "none"} {
        set iSpace $c(-imageheight)

        switch $c(-compound) {
            "top" - "bottom" {
                incr iSpace [expr {$c(-textheight)+$c(-pady)}]
            }
            "left" - "right" - "center" {
                set iSpace [expr {$iSpace > $c(-textheight) \
                        ? $iSpace : $c(-textheight)}]
            }
        }

        if {$c(-class) == "Checkbutton"} {
            set iDiam [expr {(65 * $iSpace)/100}]
        } else {
            set iDiam [expr {(75 * $iSpace)/100}]
        }
    } elseif {$c(-haveimage)} {
        set iSpace [expr {$c(-height) - 2 * $c(-borderwidth) }]
        
        if {$c(-class) == "Checkbutton"} {
            set iDiam [expr {(65 * $iSpace)/100}]
        } else {
            set iDiam [expr {(75 * $iSpace)/100}]
        }
    } elseif {$c(-havetext)} {
        set iDiam [font metrics $c(-font) -linespace]
        set iDiam [expr {($iDiam * 80)/100}]
        set iSpace [expr {[font measure $c(-font) "0"] + $iDiam}]
    } else {
        # there is no text and no image, indicator is in the center of widget.
        
        set iDiam 0
        set iSpace [expr {$c(-width)/2}] 
    }

    return [list $iSpace $iDiam]
}

proc ComputeIndicatorCoords {config} {

    array set c [dict get $config]
    
    # this is no space for indicator
 
    if {(!$c(-indicatoron)) ||
        (!($c(-idiam) > [expr {2 * $c(-borderwidth)}]))} {
        return [list -1 -1 -1 -1]
    }
    
    # compute x coordinates of indicator's border
  
    if {$c(-haveimage) && $c(-havetext)} {
        set tmp [expr {([lindex $c(-textcoords) 0] \
                - [dict get $config -textwidth]/2)}]
        # offset of text count from the left border
        set x [expr {$tmp < [lindex $c(-imagecoords) 0] \
                ? $tmp : [lindex $c(-imagecoords) 0]}]
    } elseif {$c(-haveimage)} {
        set x [lindex $c(-imagecoords) 0]
    } elseif {$c(-havetext)} {
        set x [expr {([lindex $c(-textcoords) 0] - $c(-textwidth)/2)}]
    } else {
        set x $c(-width)
    }

    if {$c(-class) == "Checkbutton"} {
        set chBtnDim 11

        # compute x and y coordinates of center of the indicator
 
        set x [expr {$x - ($c(-ispace)/2) - ($chBtnDim/2)}]
        set y [expr {$c(-y) + $c(-height)/2 - $chBtnDim/2}]
        set x1 [expr {$x + $chBtnDim - 1}]
        set y1 [expr {$y + $chBtnDim - 1}]
    } else {
        # radiobutton's indicator, it needs some graphics improvements

        set chBtnDim 9
        
        # compute x and y coordinates of center of the indicator

        set x [expr {$x - ($c(-ispace)/2) - ($chBtnDim/2)}]
        set y [expr {$c(-y) + $c(-height)/2 - $chBtnDim/2}]
        set x1 [expr {$x + $chBtnDim - 1}]
        set y1 [expr {$y + $chBtnDim - 1}]
    }
    
    return [list $x $y $x1 $y1]
}

#############################################################################
# procedure ComputeMarkCoords --
#
#       computes checkbutton's check mark coordinates. they are
#       coordinates of polygon.
#
# arguments:
#       ix  - checkbutton's indicator x coordinates
#       iy  - checkbutton's indicator x coordinates
#       ix1 - checkbutton's indicator x coordinates
#       iy1 - checkbutton's indicator x coordinates
# results:
#       coordinates of checkbutton's check mark are computed.

proc ComputeMarkCoords {ix iy ix1 iy1} {
    
    set xl [expr {$ix + 2}]
    set yl [expr {$iy + 4}]
    set xl1 $xl
    set yl1 [expr {$yl + 2}]

    set xm [expr {$ix + 1 + (0.33 * ($ix1 - $ix))}]
    set xm1 $xm
    set ym [expr {$iy1 - 2 - 2}]
    set ym1 [expr {$iy1 - 2}]

    set yp [expr {$iy + 2 + 2}]
    set yp1 [expr {$iy + 2}] 
    set xr [expr {$ix1 - 2}]
    set xr1 $xr

    return [list $xl $yl $xl1 $yl1 $xm $ym1 $xr $yp $xr1 $yp1 $xm1 $ym]
}

proc ComputeAnchor {config innerw innerh padx pady} {

    set xd 0
    set yd 0
    array set c [dict get $config]

    # horizontal part
    switch $c(-anchor) {
        center - n - s {
            set xd [expr {($c(-width) - $innerw \
                    -2*$c(-borderwidth))/2 + $c(-borderwidth)}]
        }
        w - nw - sw {
            set xd [expr {$c(-borderwidth) + $padx}]
        }
        default {
            set xd [expr {$c(-width)-$c(-borderwidth)-$padx-($innerw)}]
        }
    }

    # vertical part
    switch $c(-anchor) {
        n - ne - nw {
            set yd [expr {$pady + $c(-borderwidth)}]
        }
        w - center - e {
            set yd [expr {($c(-height) - 2*$c(-borderwidth) \
                    - $innerh)/2 + $c(-borderwidth)}]
        }
        default {
            set yd [expr {$c(-height) - $c(-borderwidth) \
                    - $pady - $innerh}]
        }
    }
    
    return [list $xd $yd]
}

#############################################################################
# proc ComputeMenuLabelSize --
#
#       computes how much space is needed by a menu's label.
#
# arguments:
#       text    - label's text.
#       font    - label's font.
#       img     - labels's image.
#       compund - label's compound.
# results:
#       a width and height required by menu's label is returned.

proc ComputeMenuLabelSize {text font img compound} {

    set width 0
    set height 0

    if {$img != ""} {
        incr width [image width $img]
        incr height [image height $img]
    }

    if {!(($img != "") && ($compound == "none"))} {
        if {$text != ""} {
            set textLayout [ComputeTextLayout $text $font 0]
            set textWidth [lindex $textLayout 0]
            set textHeight [lindex $textLayout 1]
            set linespace [font metrics $font -linespace]
            
            if {($compound != "none") && ($img != "")} {
                
                switch $compound {
                    "top" - "bottom" {
                        if {$textWidth > $width} {
                            set width $textWidth
                        }

                        incr height [expr {$linespace + 2}]
                    }
                    "left" - "right" {
                        if {$linespace > $height} {
                            set height $linespace
                        }

                        incr width [expr {$textWidth + 2}]
                    }
                    "center" {
                        if {$linespace > $height} {
                            set height $linespace
                        }
                        if {$textWidth > $width} {
                            set width $textWidth
                        }
                    }
                } 
            } else {
                set height $linespace
                set width $textWidth
            }
        } else {
            incr height $linespace
        }
    }

    incr height 1

    return [list $width $height]
}

proc ComputeMenuIndSpace {type hideMarg indOn img width height borderwidth} {
    
    set widthp $borderwidth
    set heightp 0
    
    if {(($type == "radiobutton") || ($type == "checkbutton")) \
            && ((!$hideMarg) && $indOn)} {
        
        if {$img != ""} {
            set widthp [expr {(14 * $height) / 10}]
            set heightp $height
        } else {
            set widthp $height
            set heightp $height
        }
    }

    return [list $widthp $heightp]
}

proc SetColorsToDraw {config} {

    array set c [dict get $config]

    # if widget has focus we must change it's highlightbackground color
   
    if {$c(-hasfocus)} {
        set hg $c(-highlightcolor)
    } else {
        set hg $c(-highlightbackground)
    }
    
    switch $c(-class) {
        "Labelframe" - "Text" {
            set fg $c(-foreground)
            set bg $c(-background)
        }

        "Frame" - "Canvas" - "Toplevel" - "Scrollbar" {
            set bg $c(-background)
            set fg 0
        }

        "Entry" - "Spinbox" {
            switch $c(-state) {
                "disabled" {
                    set fg $c(-disabledforeground)
                    set bg $c(-disabledbackground)
                }
                "readonly" {
                    set fg $c(-foreground)
                    set bg $c(-readonlybackground)
                }
                default {
                    set fg $c(-foreground)
                    set bg $c(-background)
                }
            }
        }

        "Button" - "Radiobutton" - "Checkbutton" - "Label" - "Listbox" \
                 - "Menubutton" - "Scale" {
            switch $c(-state) {
                "disabled" {
                    set fg $c(-disabledforeground)
                    set bg $c(-background)
                }
                "active" {
                    set fg $c(-activeforeground)
                    set bg $c(-activebackground)
                }
                default {
                    set fg $c(-foreground)
                    set bg $c(-background)
                }
            }
        }
        
        default {
            set bg $c(-background)
            set fg 0
        }
    }

    return [list $fg $bg $hg]
}

proc exportWindow {win} {

    set ::tktopdf::parentx -[winfo x $win]
    set ::tktopdf::parenty -[winfo y $win]

    set cfg [tktopdf::GetWindowConfig $win]

    ::tktocanvas::setPdfOptions \
            [expr {(2 * [dict get $cfg -highlightthickness] \
            + [dict get $cfg -width])}] \
            [expr {(2 * [dict get $cfg -highlightthickness] \
            + [dict get $cfg -height])}]

    tktocanvas::drawWindow $cfg
}

proc startFile {args} {
    
    eval ::tktocanvas::startFile $args
}

proc endFile {} {
    
    ::tktocanvas::endFile
}

proc nextPage {} {
    
    ::tktocanvas::nextPage
}

bind . <q> exit

} ;# end of namespace tktopdf
